/*
 * Accurate Z-buffer based Rendering Program for a modified NFF format
 *
 * Author:      Raghu Karinthi
 *              Department of Statistics and Computer Science
 *              Concurrent Engineering Research Center
 *              West Virginia University
 * Date:        May 26, 1994
 * Copyright (c) 1994, West Virginia University
 *
 */

#include <stdio.h>
#include <time.h>
#include <math.h>
#include "ZRendv1.h"
#include <lug.h>
#include <lugfnts.h>

void write_tga_buffer(FrameBuffer localBuff, char *filename)
{
 bitmap_hdr output;
 color *rptr, *gptr, *bptr;
 int i, j;

 allocatebitmap(&output,WINDOW_WIDTH,WINDOW_HEIGHT,COLOR_DEPTH,1<<COLOR_DEPTH);
 rptr = output.r;
 gptr = output.g;
 bptr = output.b;

   for (i=(WINDOW_WIDTH-1);i>=0;i--) 
     for (j=0;j<WINDOW_HEIGHT;j++) {
     *rptr++ = localBuff[j][i].red;
     *gptr++ = localBuff[j][i].green;
     *bptr++ = localBuff[j][i].blue;
   }
 write_tga_file(filename,&output);
}

void readTriangleData(char *string,int count,Triangle
*triangle)
{
  float x,y,z,nx,ny,nz;

  sscanf(string,"%f %f %f  %f %f %f\n",&x,&y,&z,&nx,&ny,&nz);
  triangle->vertices[count].vertex[X] = (double)x;
  triangle->vertices[count].vertex[Y] = (double)y;
  triangle->vertices[count].vertex[Z] = (double)z;
  triangle->vertices[count].normal[0] = (double)nx;
  triangle->vertices[count].normal[1] = (double)ny;
  triangle->vertices[count].normal[2] = (double)nz;
}

/* Parser for the modifief NFF file format used by this program */

void readNFFFile(char *filename, InDataSet dataset, point vrp, vector  vpn,
		 vector vupv, point prp, double *fplanep, double *bplanep,
		 double *uminp, double *umaxp, double *vminp, double *vmaxp,
		 Lights lights, BackGroundColor *backgndcolor,int *datasizep,
		 MtlProp *activeProp, int *currentlight)
{

  FILE *infile;
  char string[80];
  int state, number_sides;
  int currentTriangle;
  int count;
  float x,y,z,r,g,b,Kd,Ka,c1, c2, angle;
  double dx, dy, dz, length;
  double lx, ly, lz;
  double maxx = 0.0;
  double maxy = 0.0;
  double minx = 0.0;
  double miny = 0.0;
  point from;

  *currentlight = 0;
  currentTriangle = -1; 
  state = WAITING;
  if ((infile = fopen(filename,"r")) == NULL) {
    fprintf(stderr,"Error opening file.\n");
    exit(1);
  } else {
    while (fgets(&string[0],80,infile) != NULL) {
      if (state == WAITING) {
	switch(string[0]) {
	case 'v':
	  state = VIEWING;
	  break;
	case 'b':
	  sscanf(&string[1]," %lf %lf %lf\n",&lx,&ly,&lz);
	  backgndcolor->red = fp_fix(256*lx);
	  backgndcolor->green = fp_fix(256*ly);
	  backgndcolor->blue = fp_fix(256*lz);
	  break;
	case 'l':
	  sscanf(&string[1]," %f %f %f",&x,&y,&z);
	  lights[*currentlight].location[0] = (double)x;
	  lights[*currentlight].location[1] = (double)y;
	  lights[*currentlight].location[2] = (double)z;
	  (*currentlight)++;
	  break;
	case 'f':
	  sscanf(&string[1]," %lf %lf %lf %f %f %f %f",&lx,&ly,&lz,
		 &Kd,&Ka, &c1, &c2);
	  (*activeProp).red = fp_fix(lx);
	  (*activeProp).green = fp_fix(ly);
	  (*activeProp).blue = fp_fix(lz);
	  (*activeProp).diffuseK = (double)(Kd);
	  (*activeProp).ambientK = (double)(Ka);
	  (*activeProp).c1 = (double)c1;
	  (*activeProp).c2 = (double)c2;
	  break;
	case 'p':
	  if (string[1] == 'p') {
	    sscanf(&string[2]," %d\n",&number_sides);
	    state = POLYPAC;
	    count = 0;
	    currentTriangle++;
	  } 
	  break;
	default:
	  break;
	}
      } else if (state == VIEWING) {
	switch(string[0]) {
	case 'a':
	  if (string[1] == 't') {
	    sscanf(&string[2]," %f %f %f\n",&vrp[0],&vrp[1],&vrp[2]);
	  } else if (string[1] == 'n') {
	    sscanf(&string[6],"%f\n",&angle);
	  }
	  break;
	case 'f':
	  sscanf(&string[4]," %lf %lf %lf\n",&from[0],&from[1],&from[2]);
	  break;
	case 'h':
	  sscanf(&string[6]," %lf\n",fplanep);
	  break;
	case 'y':
	  sscanf(&string[3]," %lf\n",bplanep);
	  break;
	case 'u':
	  sscanf(&string[2]," %lf %lf %lf\n",&vupv[0],&vupv[1],&vupv[2]);
	  break;
	case 'r':
	  /* Resolution is last field.  Calculate normal to view plane */
	  dx = from[0] - vrp[0];
	  dy = from[1] - vrp[1];
	  dz = from[2] - vrp[2];
	  length = sqrt(dx*dx+dy*dy+dz*dz);
	  vpn[0] = dx/length;
	  vpn[1] = dy/length;
	  vpn[2] = dz/length;
	  *fplanep = length - *fplanep;
	  *bplanep = length - *bplanep;

	  prp[0] = 0.0;
	  prp[1] = 0.0;
	  prp[2] = length;

	  *uminp = -length * tan(angle*M_PI/360.0);
	  *vminp = *uminp;
	  *vmaxp = -*uminp;
	  *umaxp = -*uminp;
	  state = WAITING;
	  break; 
	default:
	  break; 
	}
      } else if (state == POLYPAC) {
	readTriangleData(&string[0],count,&dataset[currentTriangle]);
	dataset[currentTriangle].vertices[count].red = (fixpoint) 0;
	dataset[currentTriangle].vertices[count].green = (fixpoint) 0;
	dataset[currentTriangle].vertices[count].blue = (fixpoint) 0;
	if (count == 2) 
	  state = WAITING;
	else count++;
      }
    }    
  }
  fclose(infile);
  *datasizep = currentTriangle+1;
}


/* The following function is a modification of a function
from the SPHIGS package disributed by the Brown University */

/* ------------------------------------------------------------------
 * DESCR   :	First half of the view representation: a matrix 
 * representing the camera's orientation in 3-D world coordinates. 
 * This defines the axes of VRC.
 * --------------------------------------------------------------- */

void evaluateViewOrientationMatrix (
   point  view_ref_point,
   vector view_plane_normal,
   vector view_up_vector,
   matrix vo_matrix /* RETURN */ )
{
   vector tempvec, up_v, normal_v, proj_up_v, right_v;
   matrix trans, rot;
   double dot, tmp1;

   /* FIND THE UNIT VERSION OF THE up VECTOR */
   MAT3_COPY_VEC(up_v, view_up_vector);
   MAT3_NORMALIZE_VEC(up_v, tmp1);

   /* FIND THE UNIT VERSION OF THE view plane normal VECTOR */
   MAT3_COPY_VEC(normal_v, view_plane_normal);
   MAT3_NORMALIZE_VEC(normal_v, tmp1);

   /* Find the projection of the up vector on the film plane */
   dot = MAT3_DOT_PRODUCT(up_v, normal_v);

   MAT3_LINEAR_COMB(proj_up_v, 1.0, up_v, -dot, normal_v);
   MAT3_NORMALIZE_VEC(proj_up_v, tmp1);

   /* Find the vector that goes to the right on the film plane */
   MAT3cross_product(right_v,    normal_v, proj_up_v);

   /********* STEP 1 OF FOLEY CHAPTER VIEW.5.2 */
   /* Translate view reference point to origin */
   MAT3_SCALE_VEC(tempvec, view_ref_point, -1);
   MAT3translate(trans, tempvec);

   /********* STEP 2 OF FOLEY CHAPTER VIEW.5.2 */
   /* Rotate so normal lies on z-axis and proj_up is on y-axis */
   /* Remember: right-hand rule is in effect. */
   MAT3identity(rot);
   rot[2][0] = normal_v[X];
   rot[2][1] = normal_v[Y];
   rot[2][2] = normal_v[Z];
         
   rot[0][0] = -right_v[X];
   rot[0][1] = -right_v[Y];
   rot[0][2] = -right_v[Z];
         
   rot[1][0] = proj_up_v[X];
   rot[1][1] = proj_up_v[Y];
   rot[1][2] = proj_up_v[Z];

   /*************** PRODUCE THE FINAL PRODUCT: TRANS x ROT */
   MAT3mult(vo_matrix,  rot, trans);
}

/* The following function is a modification of a function
from the SPHIGS package disributed by the Brown University */

/* ------------------------------------------------------------------
 * DESCR : The second half of the view specification, in which the
 * 		canonical view volume is defined (NPC).
 * ---------------------------------------------------------------- */

void evaluateViewMappingMatrix (
   double umin,
   double umax,
   double vmin,
   double vmax,
   point  proj_ref_point,
   double front_plane_dist,
   double back_plane_dist,
   double viewport_minx,
   double viewport_maxx, 
   double viewport_miny,
   double viewport_maxy,
   double viewport_minz,
   double viewport_maxz,
   matrix vm_matrix  	/* RETURN */ )
{
   matrix step3_matrix, step4_matrix, step3and4_matrix, step5_matrix,
          step6_matrix, sklar_matrix;
   MAT3hvec vrp;
   MAT3vec z_axis, shear;

   MAT3_SET_HVEC (vrp,    0.0, 0.0, 0.0, 1.0);

   /**** First, we test for "garbage in" ****/
   if (proj_ref_point[Z] <= 0.0)
      fprintf (stderr, "Z-coordinate of PRP is zero or negative\n");
   if (front_plane_dist >= proj_ref_point[Z])
     fprintf (stderr, "front clip plane lies behind or on the PRP\n");
   if (back_plane_dist >= front_plane_dist)
       fprintf (stderr, "back clip plane is closer to PRP than the front clip plane\n");

   /********* STEP 3 OF FOLEY CHAPTER VIEW.5.2 */
   /* Only to be done for PERSPECTIVE. */
   /* Translate so the PRP (center of projection) is at the origin. */
   {  vector tempvec;
      MAT3_SCALE_VEC(tempvec, proj_ref_point, -1);
      MAT3translate(step3_matrix, tempvec);
      /* update vrp */
      MAT3_ADD_VEC(vrp,  vrp, tempvec);
   }

   /********* STEP 4 OF FOLEY CHAPTER VIEW.5.2 */
 /* Shear so that the center line of the view volume lies on z-axis */
   {
      vector dop_v;   /* direction of projection */

      dop_v[0] = vrp[0] + (umax+umin)/2;
      dop_v[1] = vrp[1] + (vmax+vmin)/2;
      dop_v[2] = vrp[2];
      
      MAT3_SET_VEC (z_axis, 0.0, 0.0, 1.0);
      MAT3_SET_VEC(shear, -dop_v[X] / dop_v[Z], -dop_v[Y] / dop_v[Z], 0.0);
      MAT3shear(step4_matrix, z_axis, shear);
   }

   /**** LET'S COMPUTE THE INTERMEDIATE RESULT: step3 times step4 */
   MAT3mult (step3and4_matrix, step4_matrix, step3_matrix);

   /**** STEPS 5 and 6 DIFFER BETWEEN THE TWO TYPES OF PROJECTIONS. */

   {  /********* PERSPECTIVE: STEP 5 OF FOLEY CHAPTER VIEW.5.2 */
      /* Get it to canonical perspective volume. */
      MAT3hvec vrp_prime;
      vector scale_vec;
      double zmin;

      /* FIRST, find VRP' (VRP after steps 3 and 4) */
      MAT3mult_hvec (vrp_prime, vrp, step4_matrix, FALSE);

      /* Compute a scale matrix ala Foley */
#define F front_plane_dist
#define B back_plane_dist
      MAT3_SET_VEC(scale_vec,
		   (2.0)*vrp_prime[Z]/((umax-umin)*(vrp_prime[Z]+B)),
		   (2.0)*vrp_prime[Z]/((vmax-vmin)*(vrp_prime[Z]+B)),
		   1.0/(vrp_prime[Z]+B));

      MAT3scale(step5_matrix, scale_vec);

      /************ PERSPECTIVE: STEP 6 OF FOLEY */
      /* Get it to canonical parallel volume. */
   
      zmin = (vrp_prime[Z]+F)/(vrp_prime[Z]+B);
      MAT3zero (step6_matrix);
      step6_matrix[0][0] = 1.0;
      step6_matrix[1][1] = 1.0;
      step6_matrix[2][2] = 1.0 / (1.0 - zmin);
      step6_matrix[2][3] = (0.0 - zmin) / (1.0 - zmin);
      step6_matrix[3][2] = 1.0;
   }
   /******** DO SOME MERGING */
   MAT3mult(vm_matrix,   step5_matrix, step3and4_matrix);
   MAT3mult(vm_matrix,   step6_matrix, vm_matrix);

   /********* SKLAR'S STEP */
   /* To bring the canonical (-1 to +1) to the NPC viewport.
    * This transformation will produce NDC coordinates */
   {
      matrix scalemat, transmat;
      vector scale_vec, trans_vec;
      MAT3_SET_VEC(scale_vec,
		   (viewport_maxx - viewport_minx)/2,
		   (viewport_maxy - viewport_miny)/2,
                   (viewport_maxz - viewport_minz));
      MAT3scale(scalemat,   scale_vec);
      MAT3_SET_VEC(trans_vec,
		   (viewport_minx+viewport_maxx)/2,
		   (viewport_miny+viewport_maxy)/2,
		   viewport_minz);
      MAT3translate(transmat,   trans_vec);

      MAT3mult(sklar_matrix,   transmat, scalemat);
   }
   MAT3mult(vm_matrix,   sklar_matrix, vm_matrix);
}

/*  Applying the view transform to a set of triangles */

void viewTransform(InDataSet localSet, int count, matrix transform)
{
  int i = 0;
  MAT3hvec hpoint;
  matrix itransform;

  MAT3invert(itransform,transform);
  for( i = 0; i < count; i++)
    {
      int j;
      /* transform into homogenous */
      for( j = 0; j < 3; j++)
	{
          int k;
	  for (k=0; k < 3; k++) 
	    hpoint[k] = localSet[i].vertices[j].vertex[k];
	  hpoint[3] = 1.;
	  MAT3mult_hvec(hpoint,hpoint,transform,0);
	  localSet[i].vertices[j].vertex[X] = hpoint[X]; 
	  localSet[i].vertices[j].vertex[Y] = hpoint[Y]; 
	  localSet[i].vertices[j].vertex[Z] = hpoint[Z]; 
	  for (k=0; k < 3; k++)
            hpoint[k] = localSet[i].vertices[j].normal[k];
          hpoint[3] = 1.;
          MAT3mult_hvec(hpoint,hpoint,transform,0);
          localSet[i].vertices[j].normal[X] = hpoint[X];
          localSet[i].vertices[j].normal[Y] = hpoint[Y];
          localSet[i].vertices[j].normal[Z] = hpoint[Z];
	}
    }       
}

#define TSWAP(_a_, _b_, _c_) { _c_ = _a_; _a_ = _b_; _b_ = _c_; }
#define min(_a_, _b_) ((_a_ < _b_) ? _a_ : _b_)
#define max(_a_, _b_) ((_a_ > _b_) ? _a_ : _b_)

/* The following structure is from Graphics Gems III subdirectory
accurate_scan due to Kurt Fleischer. */

struct edge {
  int ymin, ymax, xi, si;
  int r, inc, dec;
};

/* The following function is a modification of a function obtained from 
Graphics Gems III subdirectory accurate_scan due to Kurt Fleischer. */

/*  
 *  Returns the intersection of edge e with the next scanline.
 */
fixpoint FEdgeScan(struct edge *e)
{
  fixpoint x = e->xi;

  if (e->r >= 0) {
	 e->xi += e->si + 1;
	 e->r += e->dec;
  }
  else {
	 e->xi += e->si;
	 e->r += e->inc;
  }
  return x;
}

/* The following function is from Graphics Gems III subdirectory
accurate_scan due to Kurt Fleischer. */

/*	 The scan-conversion routines for coordinates at subpixel resolution.
 *
 *  SubEdgeSetup initializes the Bresenham-like scan conversion for a
 *  single edge (analogous to EdgeSetup).
 *
 *  Double lenght fixpoint is used to compute alpha below. Alpha
 *  is then truncated (this corresponds to quantizing to a 
 *  particular subpixel while computing the x-intersection of the edge
 *  with the scanline). It will not produces artifacts except for
 *  degenerate databases, in which adjacent polygons don't necessarily
 *  have coincident vertices (see Lathrop, Kirk, Voorhies, IEEE CG&A 10(5),
 *  1990 for a discussion).
 *
 */
struct edge *
SubEdgeSetup(struct edge *e, fixpoint x0, fixpoint y0,
	     fixpoint x1, fixpoint y1)
{

  fixpoint ymin, ymax, alpha, beta, sf, xi, si;
  fixpoint dx = x1-x0, dy = y1-y0;

  ymin = fp_floor(y0);
  ymax = fp_floor(y1);

  if ((dy != 0) && (ymin != ymax)) {

	 /* Alpha is related to x-intercept with scanline ymin  */
	 alpha = fp_trunc(fp_dbladd(fp_dblmultiply(fp_fraction(x0), dy),
										 fp_dblmultiply(dx, ymin - y0 + fp_fix(1.0))));
	 beta = fp_floor_div(alpha, dy);
	 xi = fp_floor(x0) + beta;

	 /* prevent overflow for v. small dy ('si' is only used if dy>=1) */
	 if (dy >= fp_fix(1.0)) {
		si = fp_floor_div(dx, dy);
		sf = dx - fp_multiply(si, dy); 

		/* (alpha - beta*dy) = fractional part of the x-intercept. */
		e->r = alpha - fp_multiply(beta, dy) + sf - dy; 
		e->si = fp_integer(si);
		e->inc = sf;
		e->dec = sf - dy;
	 }
	 e->xi = fp_integer(xi);
  }
  e->ymin = fp_integer(ymin);
  e->ymax = fp_integer(ymax);

  return(e);
}

/* The following function is a major modification of a function obtained from 
Graphics Gems III subdirectory accurate_scan due to Kurt Fleischer.
We have extended the program from 2D interpolation to 3D to be able to
interpolate z, r, g, and b. */

/*  Like integer version, but uses SubEdgeSetup */
void fix_sorted_triangle(int minv, int midv, int maxv, InDataSet tp,
			 FrameBuffer localBuff)
{
  struct edge left, right, zleft, zright, zspan, rleft, rright, rspan,
  gleft, gright, gspan, bleft, bright, bspan;
  int x, yi, xmin, xmax;
  fixpoint fxmin, fxmax, fzmin, fzmax, z, frmin, frmax, r, fgmin, fgmax, g,
  fbmin, fbmax, b;

  fixpoint x0 = fp_fix((tp->vertices[minv]).vertex[X]);
  fixpoint y0 = fp_fix((tp->vertices[minv]).vertex[Y]);
  fixpoint z0 = fp_fix((tp->vertices[minv]).vertex[Z]);
  fixpoint x1 = fp_fix((tp->vertices[midv]).vertex[X]);
  fixpoint y1 = fp_fix((tp->vertices[midv]).vertex[Y]);
  fixpoint z1 = fp_fix((tp->vertices[midv]).vertex[Z]);
  fixpoint x2 = fp_fix((tp->vertices[maxv]).vertex[X]);
  fixpoint y2 = fp_fix((tp->vertices[maxv]).vertex[Y]);
  fixpoint z2 = fp_fix((tp->vertices[maxv]).vertex[Z]);

#define r0 ((fixpoint) (tp->vertices[minv]).red)
#define g0 ((fixpoint) (tp->vertices[minv]).green)
#define b0 ((fixpoint) (tp->vertices[minv]).blue)
#define r1 ((fixpoint) (tp->vertices[midv]).red)
#define g1 ((fixpoint) (tp->vertices[midv]).green)
#define b1 ((fixpoint) (tp->vertices[midv]).blue)
#define r2 ((fixpoint) (tp->vertices[maxv]).red)
#define g2 ((fixpoint) (tp->vertices[maxv]).green)
#define b2 ((fixpoint) (tp->vertices[maxv]).blue)

  /* compute determinant using full precision (64 bits) */
  dblfixpoint aa = fp_dblmultiply(y1-y0,x2-x0);
  dblfixpoint bb = fp_dblmultiply(x1-x0,y2-y0);
  int clockwise = fp_dbllessthan(aa, bb);

  /* Setup first pair of edges */
  if (clockwise) {
	 SubEdgeSetup(&left, x0, y0, x2, y2);
	 SubEdgeSetup(&zleft, z0, y0, z2, y2);
	 SubEdgeSetup(&rleft, r0, y0, r2, y2);
	 SubEdgeSetup(&gleft, g0, y0, g2, y2);
	 SubEdgeSetup(&bleft, b0, y0, b2, y2);
	 SubEdgeSetup(&right, x0, y0, x1, y1);
	 SubEdgeSetup(&zright, z0, y0, z1, y1);
	 SubEdgeSetup(&rright, r0, y0, r1, y1);
	 SubEdgeSetup(&gright, g0, y0, g1, y1);
	 SubEdgeSetup(&bright, b0, y0, b1, y1);
  } else {
	 SubEdgeSetup(&left, x0, y0, x1, y1);
	 SubEdgeSetup(&zleft, z0, y0, z1, y1);
	 SubEdgeSetup(&rleft, r0, y0, r1, y1);
	 SubEdgeSetup(&gleft, g0, y0, g1, y1);
	 SubEdgeSetup(&bleft, b0, y0, b1, y1);
	 SubEdgeSetup(&right, x0, y0, x2, y2);
	 SubEdgeSetup(&zright, z0, y0, z2, y2);
	 SubEdgeSetup(&rright, r0, y0, r2, y2);
	 SubEdgeSetup(&gright, g0, y0, g2, y2);
	 SubEdgeSetup(&bright, b0, y0, b2, y2);
  }
  
  /* Scan first pair of edges. */
  for (yi = left.ymin + 1; yi <= min(left.ymax, right.ymax); yi++) {
	 fxmin = FEdgeScan(&left);
	 fxmax = FEdgeScan(&right);
	 fzmin = FEdgeScan(&zleft);
	 fzmax = FEdgeScan(&zright);
	 frmin = FEdgeScan(&rleft);
	 frmax = FEdgeScan(&rright);
	 fgmin = FEdgeScan(&gleft);
	 fgmax = FEdgeScan(&gright);
	 fbmin = FEdgeScan(&bleft);
	 fbmax = FEdgeScan(&bright);

	 xmin = fxmin;
	 xmax = fxmax;
	 
	 fxmin = fxmin << LOBITS;
	 fxmax = fxmax << LOBITS;
	 fzmin = fzmin << LOBITS;
	 fzmax = fzmax << LOBITS;
	 frmin = frmin << LOBITS;
	 frmax = frmax << LOBITS;
	 fgmin = fgmin << LOBITS;
	 fgmax = fgmax << LOBITS;
	 fbmin = fbmin << LOBITS;
	 fbmax = fbmax << LOBITS;  

	 SubEdgeSetup (&zspan, fzmin, fxmin, fzmax, fxmax);
	 SubEdgeSetup (&rspan, frmin, fxmin, frmax, fxmax);
	 SubEdgeSetup (&gspan, fgmin, fxmin, fgmax, fxmax);
	 SubEdgeSetup (&bspan, fbmin, fxmin, fbmax, fxmax);

	 for (x=xmin+1; x<=xmax; x++) {
	   z = FEdgeScan (&zspan);
	   r = FEdgeScan (&rspan);
	   g = FEdgeScan (&gspan);
	   b = FEdgeScan (&bspan);

	   if (localBuff[x][yi].z < z) {
	     localBuff[x][yi].z = z;
	     localBuff[x][yi].red = r;
	     localBuff[x][yi].green =g;
	     localBuff[x][yi].blue = b;
	   }
	 }
       }

  /* Setup third edge */
  if (!clockwise) {
    SubEdgeSetup(&left, x1, y1, x2, y2);
    SubEdgeSetup(&zleft, z1, y1, z2, y2);
    SubEdgeSetup(&rleft, r1, y1, r2, y2);
    SubEdgeSetup(&gleft, g1, y1, g2, y2);
    SubEdgeSetup(&bleft, b1, y1, b2, y2);
  }
  else {
    SubEdgeSetup(&right, x1, y1, x2, y2);
    SubEdgeSetup(&zright, z1, y1, z2, y2);
    SubEdgeSetup(&rright, r1, y1, r2, y2);
    SubEdgeSetup(&gright, g1, y1, g2, y2);
    SubEdgeSetup(&bright, b1, y1, b2, y2);
  }
  for (yi = max(left.ymin, right.ymin) + 1; yi <= left.ymax; yi++) {
	 fxmin = FEdgeScan(&left);
	 fxmax = FEdgeScan(&right);
	 fzmin = FEdgeScan(&zleft);
	 fzmax = FEdgeScan(&zright);
	 frmin = FEdgeScan(&rleft);
	 frmax = FEdgeScan(&rright);
	 fgmin = FEdgeScan(&gleft);
	 fgmax = FEdgeScan(&gright);
	 fbmin = FEdgeScan(&bleft);
	 fbmax = FEdgeScan(&bright);

	 xmin = fxmin;
	 xmax = fxmax;
	 fxmin = fxmin << LOBITS;
	 fxmax = fxmax << LOBITS;
	 fzmin = fzmin << LOBITS;
	 fzmax = fzmax << LOBITS;
	 frmin = frmin << LOBITS;
	 frmax = frmax << LOBITS;
	 fgmin = fgmin << LOBITS;
	 fgmax = fgmax << LOBITS;
	 fbmin = fbmin << LOBITS;
	 fbmax = fbmax << LOBITS; 

	 SubEdgeSetup (&zspan, fzmin, fxmin, fzmax, fxmax);
	 SubEdgeSetup (&rspan, frmin, fxmin, frmax, fxmax);
	 SubEdgeSetup (&gspan, fgmin, fxmin, fgmax, fxmax);
	 SubEdgeSetup (&bspan, fbmin, fxmin, fbmax, fxmax);

	 for (x=xmin+1; x<=xmax; x++) {
	   z = FEdgeScan (&zspan);
	   r = FEdgeScan (&rspan);
	   g = FEdgeScan (&gspan);
	   b = FEdgeScan (&bspan);

	   if (localBuff[x][yi].z < z) {
	     localBuff[x][yi].z = z;
	     localBuff[x][yi].red = r;
	     localBuff[x][yi].green = g;
	     localBuff[x][yi].blue = b;
	   }
	 }
       }
}

/* The following function is a modification of a function obtained from 
Graphics Gems III subdirectory accurate_scan due to Kurt Fleischer. */

/* Like integer version. */
void rasterizeTriangle(FrameBuffer localBuff, InDataSet tp)
{
  int minv=0, midv=1, maxv=2, tmp;
  if ((tp->vertices[minv]).vertex[Y] > (tp->vertices[midv]).vertex[Y])
    TSWAP(minv,midv,tmp); 
  if ((tp->vertices[minv]).vertex[Y] > (tp->vertices[maxv]).vertex[Y])
    TSWAP(minv,maxv,tmp);
  if ((tp->vertices[midv]).vertex[Y] > (tp->vertices[maxv]).vertex[Y])
    TSWAP(midv,maxv,tmp);
  fix_sorted_triangle(minv,midv,maxv,tp,localBuff);
}

/*Rasterize a number of triangles */

void rasterize(InDataSet localSet, int count, FrameBuffer localBuff)
{
  int i;
  for( i = 0; i < count; i++) 
    rasterizeTriangle(localBuff, &localSet[i]);
}

/* Compute vertex intensities using Gouraud shading; Refer to the book
by Foley, vanDam, Feiner, Hughes Chapter 16.1, 16.2 for the formulas */

void ShadeTriangles(InDataSet localSet,int *datasizep,double Iar,
	double Iag, double Iab,
        double Idr, double Idg,
	double Idb,double Kar,double Kdr,
	double Kag,double Kdg, double Kab,double Kdb, double c1, double c2,
        int num_lights, Lights lights)
{

   int i=0, j=0, k=0;
   vector temp1,temp2;
   double dist, r=0.0, g=0.0, b=0.0;
   for(i=0;i<(*datasizep);i++) {
     for (j=0; j < 3; j++) {
       r=0.0; g=0.0;b=0.0;
       for (k=0; k < num_lights; k++) {
      point p;
      p[X] = lights[k].location[X];
      p[Y] = lights[k].location[Y];
      p[Z] = lights[k].location[Z];
      MAT3_SUB_VEC(temp1,p,localSet[i].vertices[j].vertex);
      dist = sqrt(
	temp1[X]*temp1[X]+temp1[Y]*temp1[Y]+temp1[Z]*temp1[Z]);
      temp1[X] = temp1[X]/dist;
      temp1[Y] = temp1[Y]/dist;
      temp1[Z] = temp1[Z]/dist;
      dist = 1/(1 + c1*dist + c2*dist*dist);
      r += fabs((Iar*Kar+Idr*Kdr*MAT3_DOT_PRODUCT
      (temp1,localSet[i].vertices[j].normal)* dist)*MAX_INTENSITY/2);
      g += fabs((Iag*Kag +Idg*Kdg*MAT3_DOT_PRODUCT
      (temp1,localSet[i].vertices[j].normal)*dist)*MAX_INTENSITY/2);
      b += fabs((Iab*Kab + Idb*Kdb*MAT3_DOT_PRODUCT
      (temp1,localSet[i].vertices[j].normal)*dist)*MAX_INTENSITY/2);
    }
      
      localSet[i].vertices[j].red = fp_fix(r);
      localSet[i].vertices[j].green = fp_fix(g);
      localSet[i].vertices[j].blue = fp_fix(b);

    }
    }
 }

/* Main program for Z buffer rendering */

void main(int argc, char **argv)
{
  int i,j,k;
  InDataSet localSet;
  point vrp,prp;
  vector vpn, vupv;
  double fplane, bplane, umin, umax, vmin, vmax;
  int datasize;
  matrix vo_matrix, vm_matrix, transform;
  FrameBuffer localBuff;
  clock_t tm1, tm2;
  matrix result;
  MAT3hvec hpoint;
  Lights lights;
  double Iar,Iag,Iab,Idr,Idg,Idb,Kd,Ka, c1, c2;

  double minu, maxu, minv, maxv, minn, maxn;
  BackGroundColor backgndcolor;
  MtlProp activeProp;
  int num_lights =0;

  for (i=0; i < WINDOW_WIDTH; i++)
    for (j=0; j < WINDOW_HEIGHT; j++) {
      localBuff[i][j].z= 0.0;
      localBuff[i][j].red = 0;
      localBuff[i][j].green = 0;
      localBuff[i][j].blue = 0;
    }
  localSet = (InDataSet)malloc(NUM_TRIANGLES*sizeof(Triangle));

  lights = (Lights)malloc(NUM_LIGHT_SOURCES*sizeof(LightPoint));

  readNFFFile(argv[1],localSet,vrp,vpn,vupv,prp,&fplane,&bplane, &umin,&umax,
      &vmin,&vmax,lights,&backgndcolor,&datasize, &activeProp,&num_lights);
  for (i=0; i < WINDOW_WIDTH; i++)
    for (j=0; j < WINDOW_HEIGHT; j++) {
      localBuff[i][j].red = backgndcolor.red;
      localBuff[i][j].green = backgndcolor.green;
      localBuff[i][j].blue = backgndcolor.blue;
    }
    Idr = Iar = my_fp_double(activeProp.red);
    Idg = Iag = my_fp_double(activeProp.green);
    Idb = Iab = my_fp_double(activeProp.blue);
    Ka = (activeProp.ambientK);
    Kd = (activeProp.diffuseK);
    c1 = activeProp.c1;
    c2 = activeProp.c2;

  printf("Input file has %d triangles \n",datasize);
  printf("prp:%lf %lf %lf\n",prp[0],prp[1],prp[2]);
  printf("vrp:%lf %lf %lf\n",vrp[0],vrp[1],vrp[2]);
  printf("vupv:%lf %lf %lf\n",vupv[0],vupv[1],vupv[2]);
  printf("vpn:%lf %lf %lf\n",vpn[0],vpn[1],vpn[2]);
  printf("fplane: %lf\n",fplane);
  printf("bplane: %lf\n",bplane);
  printf("umin: %lf umax: %lf\n",umin,umax);
  printf("vmin: %lf vmax: %lf\n",vmin,vmax);

  printf ("Intensity is Ambient Red %lf Green %lf Blue %lf \n",Iar, Iag, Iab);
  printf ("Intensity is Diffuse Red %lf Green %lf Blue %lf \n",Idr, Idg, Idb);
  printf ("Ambient Reflection Coefficient Ka %lf \n", Ka);
  printf ("Diffuse Reflection Coefficient Kd %lf \n",  Kd);
  printf ("Attenuation function is 1/(1 + c1 * dist + c2 * dist * dist) \n");
  printf ("c1 %lf c2 %lf\n", c1, c2);
  printf ("There are %d light sources\n", num_lights);
  printf ("They are located at\n");
  for (i=0; i < num_lights; i++)
    printf ("%lf %lf %lf\n", lights[i].location[X],
	    lights[i].location[Y], lights[i].location[Z]);
  tm1 = clock ();
  ShadeTriangles(localSet,&datasize,Iar,Iag,Iab,Idr,Idg,Idb,Ka,Kd,Ka,Kd,
		Ka,Kd,c1, c2, num_lights, lights);
  tm2 = clock ();
  printf ("Shading took %f secs\n", (double) (tm2-tm1)/CLOCKS_PER_SEC);
  evaluateViewOrientationMatrix(vrp, vpn, vupv, vo_matrix);
  evaluateViewMappingMatrix(umin, umax,   vmin, vmax, prp, fplane, bplane,
		    0.0, (double) WINDOW_WIDTH, 0.0, (double) WINDOW_HEIGHT,
                    0.0,(double) WINDOW_WIDTH, vm_matrix);

  MAT3mult(transform,vm_matrix,vo_matrix);
  tm1 = clock ();
  viewTransform(localSet,datasize,transform);
  tm2 = clock ();
  printf ("Transform took %f secs\n", (double)
		(tm2-tm1)/CLOCKS_PER_SEC);
  tm1 = clock ();
  rasterize(localSet, datasize, localBuff);
  tm2 = clock ();
  printf ("Raserize took %f secs\n", (double) (tm2-tm1)/CLOCKS_PER_SEC);
  if (argc > 2) 
    write_tga_buffer(localBuff,argv[2]);
  else
    write_tga_buffer(localBuff,"a.tga");
}
