/*
 * Z-buffer based rendering program for a modified NFF format
 *
 * Author:      Raghu Karinthi
 *              Department of Statistics and Computer Science
 *              Concurrent Engineering Research Center
 *              West Virginia University
 * Date:        May 26, 1994
 * Copyright (c) 1994, West Virginia University
 *
 */

#include "matrix/macros.h"
#include "matrix/mat3.h"
#include "matrix/mat3defs.h"
#include "matrix/sph_errtypes.h"
#include "fixpoint/fixpoint.h"

/* NFF Parser constants */
#define WAITING 0
#define VIEWING 1
#define POLYPAC 3

/* Timing Constants */
#define CLOCKS_PER_SEC 1000000.0

/* 3D */
typedef MAT3vec                       point;
typedef MAT3vec                       vector;
typedef MAT3mat                       matrix;

/* PROJECTION TYPES */
#define ORTHOGRAPHIC 0
#define PERSPECTIVE  1

#define EPSILON 1e-12
#define NUM_TRIANGLES 10000 
#define WINDOW_WIDTH  500
#define WINDOW_HEIGHT 500
#define NUM_PLANES 7
#define MAX_INTENSITY 256
#define NUM_LIGHT_SOURCES 4

#define X 0
#define Y 1
#define Z 2

#define COLOR_DEPTH 24
typedef unsigned char color;
typedef fixpoint fcolor; 

typedef struct Color_Vertex_struct 
{
  point vertex;
  fcolor red, green, blue;
  vector normal;
} Color_Vertex;

typedef struct FrameBuffer_Entry_struct 
{
  color red, green, blue;
  fixpoint z;
} FrameBuffer_Entry;

typedef FrameBuffer_Entry *FrameBuffer_EntryP;

typedef struct _triangle {
  Color_Vertex vertices[3];
  unsigned char accept;
} Triangle;


typedef Triangle *InDataSet;

typedef FrameBuffer_Entry FrameBuffer[WINDOW_WIDTH][WINDOW_HEIGHT];

typedef struct BackgroundColor_struct
{
  fcolor red, green, blue;
} BackGroundColor;

typedef struct LightPoint_struct
{
  point location;
  fcolor red, green, blue;
} LightPoint;

typedef LightPoint *Lights;

typedef struct MtlProperties_struct
{
  fcolor red, green, blue;
  double diffuseK, ambientK, c1, c2;
} MtlProp;
